# Copyright: 2017, CCX Technologies
"""D-Bus Exceptions from the adbus package"""


class BusError(Exception):
    """Error Interfacing on the D-Bus

    Args:
        message: Error message
    """
    pass
